// Andrew M. Bradley ambrad@cs.stanford.edu

#ifndef INCLUDE_HD
#define INCLUDE_HD

#include <list>
#include "HmmvpUtil.hpp"

// Create an H-matrix from B. P'BQ has contiguous blocks.

namespace Hm {
  using namespace std;
  
  // Row and column global indices for a matrix block. Indexing is 1-based.
  struct Block {
    uint r0, m, c0, n;

    Block(const Vecui& p, const Vecui& q, const Vecui& r, const Vecui& c);
  };

  // Block cluster tree, represented by just the node. struct Block is a
  // leaf. Later, we make a list<Block>, and block_idx is an index into this
  // list if BNode is a leaf, else -1. struct Block is actually redundant
  // relative to a full block cluster tree, but for a matrix-vector product, the
  // tree is not necessary.
  struct BNode {
    BNode* chld[4];
    int block_idx;
    uint r0, m, c0, n;

    BNode(uint r0, uint m, uint c0, uint n);
    ~BNode();
    uint Nchldrn() const;
    uint Nnodes() const;
  };

  class Hd {
  public:
    Hd() : nblocks(0) {}
    virtual ~Hd() {}

    virtual void Permutations(Vecui& p, Vecui& q) const = 0;

    // Methods to access matrix blocks
    typedef list<Block>::const_iterator iterator;
    iterator begin() const { return blocks.begin(); }
    iterator end() const { return blocks.end(); }
    list<Block>::size_type NbrBlocks() const { return blocks.size(); }

  protected:
    // Matrix blocks
    list<Block> blocks;
    uint nblocks;
  };

  class Node;
  class HdCcCD;

  // H-matrix hierarchical decomposition, cell-centered
  class HdCc : public Hd {
  public:
    HdCc();
    virtual ~HdCc() {}

    virtual void Compute() = 0;

    void Opt_eta(double eta) throw (Exception);
    // Cluster splitting method. Default is to split a method along the major
    // axis of the cluster. Alternatively, split along axis directions only.
    enum SplitMethod { splitMajorAxis, splitAxisAligned };
    void Opt_split_method(SplitMethod sm) { split_method = sm; }
    // Method to determine whether two clusters are far enough apart. Default is
    // distPointwise, which measures the minimum pointwise distance between two
    // sets.
    enum DistanceMethod { distCentroid, distPointwise, distAxisAligned };
    void Opt_distance_method(DistanceMethod sm) { dist_method = sm; }
    
  protected:
    bool Admissible(const HdCcCD& n1, const HdCcCD& n2) const;

    // eta as used in the standard admissibility condition
    double eta;
    SplitMethod split_method;
    DistanceMethod dist_method;
  };  

  class Node;

  // Use this class if the domain (also called influence or source) points are
  // the same as the range points. These points are in the 3xN matrix A.
  class HdCcSym : public HdCc {
  public:
    HdCcSym(const Matd& A, const Matd* wgt = 0);
    virtual void Compute();
    BNode* Compute(bool want_block_cluster_tree);

    // Permutation matrix so that the blocks are contiguous. Indexing is
    // 1-based. P'BP has contiguous blocks.
    virtual void Permutations(Vecui& p, Vecui& q) const
    { p = permp; q = permp; }

  private:
    void Compute(const vector<const Node*>& ns1,
		 const vector<const Node*>& ns2);
    BNode* Compute(const Node* n1, const Node* n2);

  private:
    // Domain and range meshs
    Matd A;
    Vecui permp, permpi;
  };

  // Use this class if the domain (D) and range (R) points are different or one
  // is a subset of the other. This class returns the same results as HdCcSym
  // (with permutations p == q) if D == R; but this class takes ~25% longer in
  // this case.
  class HdCcNonsym : public HdCc {
  public:
    HdCcNonsym(const Matd& domain, const Matd& range, const Matd* wgt = 0);
    virtual void Compute();

    // Permutation matrices so that the blocks are contiguous. Indexing is
    // 1-based. P'BQ has contiguous blocks.
    virtual void Permutations(Vecui& p, Vecui& q) const
    { p = permp; q = permq; }

  private:
    void Compute(const vector<const Node*>& ns1,
		 const vector<const Node*>& ns2);

  private:
    // Domain and range meshs
    Matd D, R;
    Vecui permp, permpi, permq, permqi;
  };

}

#endif
