#include <ctype.h>
#include "mex.h"
#include "Hmat.hpp"
#include "Matrix.hpp"
using namespace std;

// Sits in memory between mex calls.
vector<Hm::Hmat*> ghms;

void Cleanup(uint id)
{
  if (id + 1 > ghms.size())
    mexErrMsgTxt("Cleanup: Invalid id.");
  if (ghms[id]) {
    delete ghms[id];
    ghms[id] = 0;
  }
  if (id == ghms.size() - 1) ghms.pop_back();
}

mxArray* Mvp
(Hm::Hmat& h, const double* x, uint m, uint ncol,
 const vector<Hm::Blint>* rs = 0, const vector<Hm::Blint>* cs = 0)
{
  if (h.GetN() != m || ncol == 0)
    mexErrMsgTxt("mvp: x is the wrong size.");

  mxArray* my = mxCreateDoubleMatrix(h.GetM(), ncol, mxREAL);
  try {
    h.Mvp(x, mxGetPr(my), ncol, rs, cs);
  } catch (Exception& e) {
    mexErrMsgTxt(e.GetMsg().c_str());
  }
  return my;
}

mxArray* MvpT
(Hm::Hmat& h, const double* x, uint m, uint ncol,
 const vector<Hm::Blint>* rs = 0, const vector<Hm::Blint>* cs = 0)
{
  if (h.GetM() != m || ncol == 0)
    mexErrMsgTxt("tmvp: x is the wrong size.");

  mxArray* my = mxCreateDoubleMatrix(h.GetN(), ncol, mxREAL);
  try {
#if 0
    if (rs) {
      if (cs) h.MvpT(x, mxGetPr(my), ncol, *rs, *cs);
      else    h.MvpT(x, mxGetPr(my), ncol, *rs);
    } else
#endif
    h.MvpT(x, mxGetPr(my), ncol);
  } catch (Exception& e) {
    mexErrMsgTxt(e.GetMsg().c_str());
  }
  return my;
}

int Init(const char *filename, uint nthreads, uint ncol, uint& nnz)
{
  if (nthreads < 0) mexErrMsgTxt("Init: Invalid nthreads.");
  if (ncol < 1) mexErrMsgTxt("Init: Invalid ncol.");

  // Get a free id
  int id = -1;
  for (uint i = 0; i < ghms.size(); i++)
    if (ghms[i] == 0) {
      id = i;
      break;
    }
  if (id == -1) {
    ghms.push_back((Hm::Hmat*)0);
    id = ghms.size() - 1;
  }

  try {
    ghms[id] = Hm::NewHmat(string(filename), ncol, nthreads);
    nnz = ghms[id]->GetNnz();
  } catch (FileException e) {
    Cleanup(id);
    mexErrMsgTxt(e.GetMsg().c_str());
  }

  return id;
}

void mexExitFunction()
{
  for (uint i = 0; i < ghms.size(); i++)
    Cleanup(i);
}

inline void GetIdxs(const mxArray* ma, vector<Hm::Blint>& rs, uint maxr)
{
  uint nrs = mxGetNumberOfElements(ma);
  const double* drs = mxGetPr(ma);
  rs.resize(nrs);
  for (uint i = 0; i < nrs; i++) {
    rs[i] = (Hm::Blint)drs[i] - 1;
    if (rs[i] < 0 || rs[i] > maxr)
      mexErrMsgTxt("Index out of bounds.");
  }
}

inline Hm::Hmat& GetHmat(const mxArray* ma)
{
  uint id = (uint)mxGetScalar(ma);
  if (ghms.empty() || id > ghms.size() - 1 || ghms[id] == 0)
    mexErrMsgTxt("mvp: Invalid id.");
  return *ghms[id];
}

// Implements
//   [id nnz] = hm_mvp('init',filename,[nthreads],[ncol])
//   hm_mvp('cleanup',id)
//   y = hm_mvp('mvp',id,x,[rs],[cs])
//     rs,cs are optional subsets of rows, cols. indexing starts at 1. Just rs
//     can be specified. See Hmat.hpp for more details.
//   y = hm_mvp('tmvp',id,x,[rs],[cs])
//     Transpose MVP.
//   nthreads = hm_mvp('threads',id,nthreads);
//   hm_mvp('stsave',id);
//   hm_mvp('strelease',id);
//   hm_mvp('e2bon',id);
//   hm_mvp('e2boff',id);
//   hm_mvp('pon',id);
//   hm_mvp('poff',id);
//   nf2 = hm_mvp('fronorm2',id)
//   n1 = hm_mvp('norm1',id)
//   Brc = hm_mvp('extract',id,rs,cs).
//   [I J S] = hm_mvp('bblocks',id,[cutoff]);
//   m/n = hm_mvp('getm/n',id);
void mexFunction(int nlhs, mxArray** plhs, int nrhs, const mxArray** prhs)
{
  // Clean up command.
  int strlen = mxGetNumberOfElements(prhs[0]) + 1;
  vector<char> vfn(strlen);
  mxGetString(prhs[0], &vfn[0], strlen);
  for (int i = 0; i < strlen - 1; i++) vfn[i] = tolower(vfn[i]);
  string fn(&vfn[0]);

  // init
  if (fn[0] == 'i' || fn == "init") {
    mexAtExit(&mexExitFunction);
    if (nrhs < 2 || nrhs > 4 || nlhs < 1 || !mxIsChar(prhs[1]))
      mexErrMsgTxt("[id nnz] = hm_mvp('init',filename,[nthreads],[ncol])");
    uint n = mxGetNumberOfElements(prhs[1]) + 1;
    char* filename = new char[n];
    mxGetString(prhs[1], filename, n);
    uint nthreads = 1;
    if (nrhs > 2) {
      if (!mxIsDouble(prhs[2]) || mxGetNumberOfElements(prhs[2]) != 1)
	mexErrMsgTxt("nthreads should be a positive integer");
      nthreads = (int)mxGetScalar(prhs[2]);
    }
    uint ncol = 1;
    if (nrhs > 3) {
      if (!mxIsDouble(prhs[3]) || mxGetNumberOfElements(prhs[3]) != 1)
	mexErrMsgTxt("ncol should be a positive integer");
      ncol = (int)mxGetScalar(prhs[3]);
    }
    uint nnz;
    uint id = Init(filename, nthreads, ncol, nnz);
    delete[] filename;
    plhs[0] = mxCreateDoubleScalar((double)id);
    if (nlhs > 1) plhs[1] = mxCreateDoubleScalar((double)nnz);
  }

  // cleanup
  else if (fn[0] == 'c' || fn == "cleanup") {
    if (nrhs != 2 || !mxIsDouble(prhs[1]))
      mexErrMsgTxt("hm_mvp('cleanup',id)");
    uint id = (int)mxGetScalar(prhs[1]);
    Cleanup(id);
  }

  // mvp
  else if (fn[0] == 'm' || fn == "mvp") {
    if (nrhs < 3 || nrhs > 5 || nlhs != 1 || !mxIsDouble(prhs[1]) ||
	!mxIsDouble(prhs[2]) || (nrhs >= 4 && !mxIsDouble(prhs[3])) ||
	(nrhs == 5 && !mxIsDouble(prhs[4])))
      mexErrMsgTxt("y = hm_mvp('mvp',id,x,[rs],[cs])");
    Hm::Hmat& h = GetHmat(prhs[1]);
    const mxArray* mx = prhs[2];
    vector<Hm::Blint> *prs = NULL, *pcs = NULL;
    vector<Hm::Blint> rs, cs;
    if (nrhs >= 4) {
      GetIdxs(prhs[3], rs, h.GetM() - 1);
      if (!rs.empty()) prs = &rs;
      if (nrhs == 5) {
        GetIdxs(prhs[4], cs, h.GetN() - 1);
        if (!cs.empty()) pcs = &cs;
      }
    }
    plhs[0] = Mvp(h, mxGetPr(mx), mxGetM(mx), mxGetN(mx), prs, pcs);
  }

  // tranpose mvp
  else if (fn == "tmvp") {
    if (nrhs < 3 || nrhs > 5 || nlhs != 1 || !mxIsDouble(prhs[1]) ||
	!mxIsDouble(prhs[2]) || (nrhs >= 4 && !mxIsDouble(prhs[3])) ||
	(nrhs == 5 && !mxIsDouble(prhs[4])))
      mexErrMsgTxt("y = hm_mvp('tmvp',id,x,[rs],[cs])");
    Hm::Hmat& h = GetHmat(prhs[1]);
    const mxArray* mx = prhs[2];
    vector<Hm::Blint> *prs = NULL, *pcs = NULL;
    vector<Hm::Blint> rs, cs;
    if (nrhs >= 4) {
      GetIdxs(prhs[3], rs, h.GetM() - 1);
      if (!rs.empty()) prs = &rs;
      if (nrhs == 5) {
        GetIdxs(prhs[4], cs, h.GetN() - 1);
        if (!cs.empty()) pcs = &cs;
      }
    }
    plhs[0] = MvpT(h, mxGetPr(mx), mxGetM(mx), mxGetN(mx), prs, pcs);
  }

  // Frobenius norm
  else if (fn == "fronorm2") {
    if (nlhs != 1 || nrhs > 2 || !mxIsDouble(prhs[1]) ||
	mxGetNumberOfElements(prhs[1]) != 1)
      mexErrMsgTxt("nf2 = hm_mvp('fronorm2',id)");
    Hm::Hmat& h = GetHmat(prhs[1]);
    plhs[0] = mxCreateDoubleScalar(h.NormFrobenius2());
  }

  // 1-norm
  else if (fn == "norm1") {
    if (nlhs != 1 || nrhs > 2 || !mxIsDouble(prhs[1]) ||
	mxGetNumberOfElements(prhs[1]) != 1)
      mexErrMsgTxt("nf2 = hm_mvp('norm1',id)");
    Hm::Hmat& h = GetHmat(prhs[1]);
    plhs[0] = mxCreateDoubleScalar(h.NormOne());
  }

  // State
  else if (fn.substr(0, 2) == "st") {
    if (nrhs < 2 || mxGetNumberOfElements(prhs[0]) < 3)
      mexErrMsgTxt("mvp: Invalid function");
    Hm::Hmat& h = GetHmat(prhs[1]);
    switch (fn[2]) {
    case 's':
      h.SaveState();
      break;
    case 'r':
      h.ReleaseState();
      break;
    }
  }

  // Permutations
  else if (fn.substr(0, 2) == "po") {
    if (nrhs < 2 || mxGetNumberOfElements(prhs[0]) < 3)
      mexErrMsgTxt("mvp: Invalid function");
    Hm::Hmat& h = GetHmat(prhs[1]);
    switch (fn[2]) {
    case 'n': case 'N':
      h.TurnOnPermute();
      break;
    case 'f': case 'F':
      h.TurnOffPermute();
      break;
    }
  }

  // Element-to-block map
  else if (fn.substr(0, 4) == "e2bo") {
    if (nrhs < 2 || mxGetNumberOfElements(prhs[0]) < 5)
      mexErrMsgTxt("mvp: Invalid function");
    Hm::Hmat& h = GetHmat(prhs[1]);
    switch (fn[4]) {
    case 'n': case 'N':
      h.TurnOnElementToBlockMap();
      break;
    case 'f': case 'F':
      h.TurnOffElementToBlockMap();
      break;
    }    
  }

  // Extract
  else if (fn == "extract") {
    if (nlhs != 1 || nrhs != 4 || !mxIsDouble(prhs[1]) ||
	mxGetNumberOfElements(prhs[1]) != 1)
      mexErrMsgTxt("Brc = hm_mvp('extract',id,rs,cs)");
    Hm::Hmat& h = GetHmat(prhs[1]);
    vector<Hm::Blint> rs, cs;
    GetIdxs(prhs[2], rs, h.GetM() - 1);
    GetIdxs(prhs[3], cs, h.GetN() - 1);
    plhs[0] = mxCreateDoubleMatrix(rs.size(), cs.size(), mxREAL);
    h.Extract(rs, cs, mxGetPr(plhs[0]));
  }

  // [I J S] triple of full-block blocks
  else if (fn == "bblocks") {
    if (nlhs != 3 || nrhs < 2 || !mxIsDouble(prhs[1]) ||
	mxGetNumberOfElements(prhs[1]) != 1)
      mexErrMsgTxt("[I J S] = hm_mvp('bblocks',id,[cutoff])");
    Hm::Hmat& h = GetHmat(prhs[1]);
    double cutoff = -1;
    if (nrhs > 2) cutoff = (double) mxGetScalar(prhs[2]);
    vector<Hm::Blint> I, J;
    vector<double> S;
    h.FullBlocksIJS(I, J, S, cutoff);
    size_t n = I.size();
    plhs[0] = mxCreateDoubleMatrix(n, 1, mxREAL);
    double* p = mxGetPr(plhs[0]);
    for (size_t i = 0; i < n; i++) p[i] = (double) I[i];
    I.clear();
    plhs[1] = mxCreateDoubleMatrix(n, 1, mxREAL);
    p = mxGetPr(plhs[1]);
    for (size_t i = 0; i < n; i++) p[i] = (double) J[i];
    J.clear();
    plhs[2] = mxCreateDoubleMatrix(n, 1, mxREAL);
    p = mxGetPr(plhs[2]);
    for (size_t i = 0; i < n; i++) p[i] = (double) S[i];
    S.clear();
  }

  // Number of threads
  else if (fn == "threads") {
    if (nlhs != 1 || nrhs != 3 || mxGetNumberOfElements(prhs[1]) != 1)
      mexErrMsgTxt("nthreads = hm_mvp('threads',id,nthreads)");
    Hm::Hmat& h = GetHmat(prhs[1]);
    int nthreads = (int) mxGetScalar(prhs[2]);
    if (nthreads < 1)
      mexErrMsgTxt("nthreads must be > 0");
    nthreads = h.SetThreads(nthreads);
    plhs[0] = mxCreateDoubleScalar((double) nthreads);
  }

  else if (fn == "getm" || fn == "getn") {
    if (nlhs != 1 || nrhs > 2 || !mxIsDouble(prhs[1]) ||
	mxGetNumberOfElements(prhs[1]) != 1)
      mexErrMsgTxt("m/n = hm_mvp('getm/n',id)");
    Hm::Hmat& h = GetHmat(prhs[1]);
    if (fn[3] == 'm')
      plhs[0] = mxCreateDoubleScalar(h.GetM());
    else
      plhs[0] = mxCreateDoubleScalar(h.GetN());
  }

  else if (fn == "nnz") {
    if (nlhs != 1 || nrhs != 4 || !mxIsDouble(prhs[1]) ||
	mxGetNumberOfElements(prhs[1]) != 1)
      mexErrMsgTxt("nnz = hm_mvp('nnz',id,rs,cs)");
    Hm::Hmat& h = GetHmat(prhs[1]);
    vector<Hm::Blint> rs, cs;
    GetIdxs(prhs[2], rs, h.GetM() - 1);
    GetIdxs(prhs[3], cs, h.GetN() - 1);
    Hm::Blint nnz = h.GetNnz(rs, cs);
    plhs[0] = mxCreateDoubleScalar((double) nnz);
  }
  
  else {
    mexErrMsgTxt((string("Invalid function: ") + fn).c_str());
  }
}
